/*! \file clu_subs.h
 *
 * \brief C++ porting of CLU functions and subroutines.
 *
 * Remember that FORTRAN passes arguments by reference, so, every time we use
 * a subroutine call, we need to add a referencing layer to the C++ variable.
 * All the functions defined below need to be properly documented and ported
 * to C++.
 *
 * Currently, only basic documenting information about functions and parameter
 * types are given, to avoid doxygen warning messages.
 */

#ifndef INCLUDE_COMMONS_H_
#include "Commons.h"
#endif

#ifndef INCLUDE_CLU_SUBS_H_
#define INCLUDE_CLU_SUBS_H_

#include <complex>

// >>> DECLARATION OF SPH_SUBS <<<
extern void aps(double ****zpv, int li, int nsph, C1 *c1, double sqk, double *gaps);
extern std::complex<double> dconjg(std::complex<double> value);
extern void dme(
		int li, int i, int npnt, int npntts, double vk, double exdc, double exri,
		C1 *c1, C2 *c2, int &jer, int &lcalc, std::complex<double> &arg
);
extern void rabas(
		int inpol, int li, int nsph, C1 *c1, double **tqse, std::complex<double> **tqspe,
		double **tqss, std::complex<double> **tqsps
);
extern void rbf(int n, double x, int &nm, double sj[]);
extern void rnf(int n, double x, int &nm, double sy[]);
extern void sphar(double cth, double sth, double cph, double sph, int lm, std::complex<double> *ylm);
extern void thdps(int lm, double ****zpv);
// >>> END OF SPH_SUBS DECLARATION <<<
void logmat(std::complex<double> **mat, int rows, int cols);

/*! \brief C++ porting of CDTP
 *
 * \param z: `complex<double>`
 * \param am: Matrix of complex.
 * \param i: `int`
 * \param jf: `int`
 * \param k: `int`
 * \param nj: `int`
 */
std::complex<double> cdtp(
		std::complex<double> z, std::complex<double> **am, int i, int jf,
		int k, int nj
) {
	/* NOTE: the original FORTRAN code treats the AM matrix as a
	 * vector. This is not directly allowed in C++ and it requires
	 * accounting for the different dimensions.
	 */
	std::complex<double> result = z;
	if (nj > 0) {
		int jl = jf + nj - 1;
		for (int j = jf; j <= jl; j++) {
			result += (am[i - 1][j - 1] * am[j - 1][k - 1]);
		}
	}
	return result;
}

/*! \brief C++ porting of CGEV
 *
 * \param ipamo: `int`
 * \param mu: `int`
 * \param l: `int`
 * \param m: `int`
 * \return result: `double`
 */
double cgev(int ipamo, int mu, int l, int m) {
	double result = 0.0;
	double xd = 0.0, xn = 0.0;
	if (ipamo == 0) {
		if (m != 0 || mu != 0) { // label 10
			if (mu != 0) {
				xd = 2.0 * l * (l + 1);
				if (mu <= 0) {
					xn = 1.0 * (l + m) * (l - m + 1);
					result = sqrt(xn / xd);
				} else { // label 15
					xn = 1.0 * (l - m) * (l + m + 1);
					result = -sqrt(xn / xd);
				}
			} else { // label 20
				xd = 1.0 * (l + 1) * l;
				xn = -1.0 * m;
				result = xn / sqrt(xd);
			}
		}
	} else { // label 30
		xd = 2.0 * l * (l * 2 - 1);
		if (mu < 0) { // label 35
			xn = 1.0 * (l - 1 + m) * (l + m);
		} else if (mu == 0) { // label 40
			xn = 2.0 * (l - m) * (l + m);
		} else { // mu > 0, label 45
			xn = 1.0 * (l - 1 - m) * (l - m);
		}
		result = sqrt(xn / xd);
	}
	return result;
}

/*! \brief C++ porting of R3JJR
 *
 * \param j2: `int`
 * \param j3: `int`
 * \param m2: `int`
 * \param m3: `int`
 * \param c6: `C6 *`
 */
void r3jjr(int j2, int j3, int m2, int m3, C6 *c6) {
	int jmx = j3 + j2;
	int jdf = j3 - j2;
	int m1 = -m2 - m3;
	int abs_jdf = (jdf >= 0) ? jdf : -jdf;
	int abs_m1 = (m1 >= 0) ? m1 : -m1;
	int jmn = (abs_jdf > abs_m1) ? abs_jdf : abs_m1;
	int njmo = jmx - jmn;
	int jf = jmx + jmx + 1;
	int isn = 1;
	if ((jdf + m1) % 2 != 0) isn = -1;
	if (njmo <= 0) {
		double sj = 1.0 * jf;
		double cnr = (1.0 / sqrt(sj)) * isn;
		c6->rac3j[0] = cnr;
	} else { // label 15
		double sjt = 1.0;
		double sjr = 1.0 * jf;
		int jsmpos = (jmx + 1) * (jmx + 1);
		int jdfs = jdf * jdf;
		int m1s = m1 * m1;
		int mdf = m3 - m2;
		int idjc = m1 * (j3 * (j3 + 1) - j2 * (j2 +1));
		int j1 = jmx;
		int j1s = j1 * j1;
		int j1po = j1 + 1;
		double ccj = 1.0 * (j1s - jdfs) * (j1s - m1s);
		double cj = sqrt(ccj * (jsmpos - j1s));
		double dj = 1.0 * jf * (j1 * j1po * mdf + idjc);
		// In old version, CJP was defined here. Did not work.
		// double cjp = 0.0
		if (njmo <= 1) {
			c6->rac3j[0] = -dj / (cj * j1po);
			double sj = sjr + (c6->rac3j[0] * c6->rac3j[0]) * (jf - 2);
			double cnr = (1.0 / sqrt(sj)) * isn;
			c6->rac3j[1] = cnr;
			c6->rac3j[0] *= cnr;
		} else { // label 20
			double cjp = 0.0;
			int nj = njmo + 1;
			int nmat = (nj + 1) / 2;
			c6->rac3j[nj - 1] = 1.0;
			c6->rac3j[njmo - 1] = -dj / (cj * j1po);
			if (nmat != njmo) {
				int nbr = njmo - nmat;
				for (int ibr45 = 1; ibr45 <= nbr; ibr45++) {
					int irr = nj - ibr45;
					jf -= 2;
					j1--;
					j1s = j1 * j1;
					j1po = j1 + 1;
					cjp = cj;
					ccj = 1.0 * (j1s - jdfs) * (j1s - m1s);
					cj = sqrt(ccj * (jsmpos - j1s));
					sjt = c6->rac3j[irr - 1] * c6->rac3j[irr - 1];
					dj = 1.0 * jf * (j1 * j1po * mdf + idjc);
					c6->rac3j[irr - 2] = -(c6->rac3j[irr - 1] * dj
							+ c6->rac3j[irr] * cjp * j1) / (cj * j1po);
					sjr += (sjt * jf);
				} // ibr45 loop
			}
			// label 50
			double osjt = sjt;
			sjt = c6->rac3j[nmat - 1] * c6->rac3j[nmat - 1];
			if (sjt >= osjt) {
				sjr += (sjt * (jf - 2));
			} else { // label 55
				nmat++;
			}
			// label 60
			double racmat = c6->rac3j[nmat - 1];
			c6->rac3j[0] = 1.0;
			jf = jmn + jmn + 1;
			double sjl = 1.0 * jf;
			j1 = jmn;
			if (j1 != 0) {
				j1po = j1 + 1;
				int j1pos = j1po * j1po;
				double ccjp = 1.0 * (j1pos - jdfs) * (j1pos - m1s);
				cjp = sqrt(ccjp * (jsmpos - j1pos));
				dj = 1.0 * jf * (j1 * j1po * mdf + idjc);
				c6->rac3j[1] = - dj / (cjp * j1);
			} else { // label 62
				cjp = sqrt(1.0 * (jsmpos - 1));
				dj = 1.0 * mdf;
				c6->rac3j[1] = -dj / cjp;
			}
			// label 63
			int nmatmo = nmat - 1;
			if (nmatmo >= 2) {
				for (int irl70 = 2; irl70 <= nmatmo; irl70++) {
					jf += 2;
					j1++;
					j1po = j1 + 1;
					int j1pos = j1po * j1po;
					cj = cjp;
					double ccjp = 1.0 * (j1pos - jdfs) * (j1pos - m1s);
					cjp = sqrt(ccjp * (jsmpos - j1pos));
					sjt = c6->rac3j[irl70 - 1] * c6->rac3j[irl70 - 1];
					dj = 1.0 * jf * (j1 * j1po * mdf + idjc);
					c6->rac3j[irl70] = -(
							c6->rac3j[irl70 - 1] * dj
							+ c6->rac3j[irl70 - 2] * cj * j1po
					) / (cjp * j1);
					sjl += (sjt * jf);
				}
			}
			// label 75
			double ratrac = racmat / c6->rac3j[nmat - 1];
			double rats = ratrac * ratrac;
			double sj = sjr + sjl * rats;
			c6->rac3j[nmat - 1] = racmat;
			double cnr = (1.0 / sqrt(sj)) * isn;
			for (int irr80 = nmat; irr80 <= nj; irr80++) c6->rac3j[irr80 - 1] *= cnr;
			double cnl = cnr * ratrac;
			for (int irl85 = 1; irl85 <= nmatmo; irl85++) c6->rac3j[irl85 - 1] *= cnl;
		}
	}
}


/*! \brief C++ porting of GHIT
 *
 * \param ihi: `int`
 * \param ipamo: `int`
 * \param nbl: `int`
 * \param l1: `int`
 * \param m1: `int`
 * \param l2: `int`
 * \param m2: `int`
 * \param c1: `C1 *`
 * \param c1ao: `C1_AddOns *`
 * \param c4: `C4 *`
 * \param c6: `C6 *`
 */
std::complex<double> ghit(
		int ihi, int ipamo, int nbl, int l1, int m1, int l2, int m2, C1 *c1,
		C1_AddOns *c1ao, C4 *c4, C6 *c6
) {
	/* NBL identifies transfer vector going from N2 to N1;
	 * IHI=0 for Hankel, IHI=1 for Bessel, IHI=2 for Bessel from origin;
	 * depending on IHI, IPAM=0 gives H or I, IPAM= 1 gives K or L. */
	const std::complex<double> cc0(0.0, 0.0);
	const std::complex<double> uim(0.0, 1.0);
	std::complex<double> csum(0.0, 0.0), cfun(0.0, 0.0);
	std::complex<double> result = cc0;

	if (ihi == 2) {
		if (c1->rxx[nbl - 1] == 0.0 && c1->ryy[nbl - 1] == 0.0 && c1->rzz[nbl - 1] == 0.0) {
			if (ipamo == 0) {
				if (l1 == l2 && m1 == m2) result = std::complex(1.0, 0.0);
			}
			return result;
		}
	}
	// label 10
	int l1mp = l1 - ipamo;
	int l1po = l1 + 1;
	int m1mm2 = m1 - m2;
	int m1mm2m = (m1mm2 > 0) ? m1mm2 + 1 : 1 - m1mm2;
	int lminpo = (l2 - l1mp > 0) ? l2 - l1mp + 1 : l1mp - l2 + 1;
	int lmaxpo = l2 + l1mp + 1;
	int i3j0in = c1ao->ind3j[l1mp][l2 - 1];
	int ilin = -1;
	if (m1mm2m > lminpo && (m1mm2m - lminpo) % 2 != 0) ilin = 0;
	int isn = 1;
	if (m1 % 2 != 0) isn *= -1;
	if (lminpo % 2 == 0) {
		isn *= -1;
		if (l2 > l1mp) isn *= -1;
	}
	// label 12
	int nblmo = nbl - 1;
	if (ihi != 2) {
		int nbhj = nblmo * c4->litpo;
		int nby = nblmo * c4->litpos;
		if (ihi != 1) {
			for (int jm24 = 1; jm24 <= 3; jm24++) {
				csum = cc0;
				int mu = jm24 - 2;
				int mupm1 = mu + m1;
				int mupm2 = mu + m2;
				if (mupm1 >= -l1mp && mupm1 <= l1mp && mupm2 >= - l2 && mupm2 <= l2) {
					int jsn = -isn;
					if (mu == 0) jsn = isn;
					double cr = cgev(ipamo, mu, l1, m1) * cgev(0, mu, l2, m2);
					int i3j0 = i3j0in;
					if (mupm1 == 0 && mupm2 == 0) {
						int lt14 = lminpo;
						while (lt14 <= lmaxpo) {
							i3j0++;
							int l3 = lt14 - 1;
							int ny = l3 * l3 + lt14;
							double aors = 1.0 * (l3 + lt14);
							double f3j = (c1ao->v3j0[i3j0 - 1] * c1ao->v3j0[i3j0 - 1] * sqrt(aors)) * jsn;
							cfun = (c1ao->vh[nbhj + lt14 - 1] * c1ao->vyhj[nby + ny - 1]) * f3j;
							csum += cfun;
							jsn *= -1;
							lt14 += 2;
						}
						// goes to 22
					} else { // label 16
						r3jjr(l1mp, l2, -mupm1, mupm2, c6);
						int il = ilin;
						int lt20 = lminpo;
						while (lt20 <= lmaxpo) {
							i3j0++;
							if (m1mm2m <= lt20) {
								il += 2;
								int l3 = lt20 - 1;
								int ny = l3 * l3  + lt20 + m1mm2;
								double aors = 1.0 * (l3 + lt20);
								double f3j = (c6->rac3j[il - 1] * c1ao->v3j0[i3j0 - 1] * sqrt(aors)) * jsn;
								//printf("DEBUG: VH( %d ) = (%lE, %lE)\n", (nbhj + lt20), c1ao->vh[nbhj + lt20 - 1].real(), c1ao->vh[nbhj + lt20 - 1].imag());
								//printf("DEBUG: VYHJ( %d ) = (%lE, %lE)\n", (nby + ny), c1ao->vyhj[nby + ny - 1].real(), c1ao->vyhj[nby + ny - 1].imag());
								cfun = (c1ao->vh[nbhj + lt20 - 1] * c1ao->vyhj[nby + ny - 1]) * f3j;
								csum += cfun; // we were here
							}
							// label 20
							jsn *= -1;
							lt20 += 2;
						}
					}
					// label 22
					csum *= cr;
					result += csum;
				}
				// Otherwise there is nothing to add
			} // jm24 loop. Should go to 70
		} else { // label 30, IHI == 1
			for (int jm44 = 1; jm44 <= 3; jm44++) {
				csum = cc0;
				int mu = jm44 - 2;
				int mupm1 = mu + m1;
				int mupm2 = mu + m2;
				if (mupm1 >= -l1mp && mupm1 <= l1mp && mupm2 >= - l2 && mupm2 <= l2) {
					int jsn = - isn;
					if (mu == 0) jsn = isn;
					double cr = cgev(ipamo, mu, l1, m1) * cgev(0, mu, l2, m2);
					int i3j0 = i3j0in;
					if (mupm1 == 0 && mupm2 == 0) {
						int lt34 = lminpo;
						while (lt34 <= lmaxpo) {
							i3j0++;
							int l3 = lt34 - 1;
							int ny = l3 * l3 + lt34;
							double aors = 1.0 * (l3 + lt34);
							double f3j = (c1ao->v3j0[i3j0 - 1] * c1ao->v3j0[i3j0 - 1] * sqrt(aors)) * jsn;
							cfun = (c1ao->vh[nbhj + lt34 - 1] * c1ao->vyhj[nby + ny - 1]) * f3j;
							csum += cfun;
							jsn *= -1;
							lt34 += 2;
						}
						// goes to 42
					} else { // label 36
						r3jjr(l1mp, l2, -mupm1, mupm2, c6);
						int il = ilin;
						int lt40 = lminpo;
						while (lt40 <= lmaxpo) {
							i3j0++;
							if (m1mm2m <= lt40) {
								il += 2;
								int l3 = lt40 - 1;
								int ny = l3 * l3  + lt40 + m1mm2;
								double aors = 1.0 * (l3 + lt40);
								double f3j = (c6->rac3j[il - 1] * c1ao->v3j0[i3j0 - 1] * sqrt(aors)) * jsn;
								cfun = (c1ao->vh[nbhj + lt40 - 1] * c1ao->vyhj[nby + ny - 1]) * f3j;
								csum += cfun;
							}
							// label 40
							jsn *= -1;
							lt40 += 2;
						}
					}
					// label 42
					csum *= cr;
					result += csum;
				}
				// Otherwise there is nothing to add
			} // jm44 loop. Should go to 70
		}
		// goes to 70
	} else { // label 50, IHI == 2
		int nbhj = nblmo * c4->lmtpo;
		int nby = nblmo * c4->lmtpos;
		for (int jm64 = 1; jm64 <= 3; jm64++) {
			csum = cc0;
			int mu = jm64 - 2;
			int mupm1 = mu + m1;
			int mupm2 = mu + m2;
			if (mupm1 >= -l1mp && mupm1 <= l1mp && mupm2 >= - l2 && mupm2 <= l2) {
				int jsn = -isn;
				if (mu == 0) jsn = isn;
				double cr = cgev(ipamo, mu, l1, m1) * cgev(0, mu, l2, m2);
				int i3j0 = i3j0in;
				if (mupm1 == 0 && mupm2 == 0) {
					int lt54 = lminpo;
					while (lt54 <= lmaxpo) {
						i3j0++;
						int l3 = lt54 - 1;
						int ny = l3 * l3 + lt54;
						double aors = 1.0 * (l3 + lt54);
						double f3j = (c1ao->v3j0[i3j0 - 1] * c1ao->v3j0[i3j0 - 1] * sqrt(aors)) * jsn;
						cfun = (c1ao->vj0[nbhj + lt54 - 1] * c1ao->vyj0[nby + ny - 1]) * f3j;
						csum += cfun;
						jsn *= -1;
						lt54 += 2;
					}
					// goes to 62
				} else { // label 56
					r3jjr(l1mp, l2, -mupm1, mupm2, c6);
					int il = ilin;
					int lt60 = lminpo;
					while (lt60 <= lmaxpo) {
						i3j0++;
						if (m1mm2m <= lt60) {
							il += 2;
							int l3 = lt60 - 1;
							int ny = l3 * l3  + lt60 + m1mm2;
							double aors = 1.0 * (l3 + lt60);
							double f3j = (c6->rac3j[il - 1] * c1ao->v3j0[i3j0 - 1] * sqrt(aors)) * jsn;
							cfun = (c1ao->vj0[nbhj + lt60 - 1] * c1ao->vyj0[nby + ny - 1]) * f3j;
							csum += cfun;
						}
						// label 60
						jsn *= -1;
						lt60 += 2;
					}
				}
				// label 62
				csum *= cr;
				result += csum;
			}
			// Otherwise there is nothing to add
		} // jm64 loop. Should go to 70
	}
	// label 70
	const double four_pi = acos(0.0) * 8.0;
	if (ipamo != 1) {
		double cr = sqrt(four_pi * (l1 + l1po) * (l2 + l2 + 1));
		result *= cr;
	} else {
		double cr = sqrt(four_pi * (l1 + l1mp) * (l1 + l1po) * (l2 + l2 + 1) / l1po);
		result *= (cr * uim);
	}
	return result;
}

/*! \brief C++ porting of CMS
 *
 * \param am: Matrix of complex.
 * \param c1: `C1 *`
 * \param c1ao: `C1_AddOns *`
 * \param c4: `C4 *`
 * \param c6: `C6 *`
 */
void cms(std::complex<double> **am, C1 *c1, C1_AddOns *c1ao, C4 *c4, C6 *c6) {
	std::complex<double> dm, de, cgh, cgk;
	const std::complex<double> cc0(0.0, 0.0);
	int ndi = c4->nsph * c4->nlim;
	int nbl = 0;
	int nsphmo = c4->nsph - 1;
	for (int n1 = 1; n1 <= nsphmo; n1++) { // GPU portable?
		int in1 = (n1 - 1) * c4->nlim;
		int n1po = n1 + 1;
		for (int n2 = n1po; n2 <= c4->nsph; n2++) {
			int in2 = (n2 - 1) * c4->nlim;
			nbl++;
			for (int l1 = 1; l1 <= c4->li; l1++) {
				int l1po = l1 + 1;
				int il1 = l1po * l1;
				int l1tpo = l1po + l1;
				for (int im1 = 1; im1 <= l1tpo; im1++) {
					int m1 = im1 - l1po;
					int ilm1 = il1 + m1;
					int ilm1e = ilm1 + ndi;
					int i1 = in1 + ilm1;
					int i1e = in1 + ilm1e;
					int j1 = in2 + ilm1;
					int j1e = in2 + ilm1e;
					for (int l2 = 1; l2 <= c4->li; l2++) {
						int l2po = l2 + 1;
						int il2 = l2po * l2;
						int l2tpo = l2po + l2;
						int ish = ((l2 + l1) % 2 == 0) ? 1 : -1;
						int isk = -ish;
						for (int im2 = 1; im2 <= l2tpo; im2++) {
							int m2 = im2 - l2po;
							int ilm2 = il2 + m2;
							int ilm2e = ilm2 + ndi;
							int i2 = in2 + ilm2;
							int i2e = in2 + ilm2e;
							int j2 = in1 + ilm2;
							int j2e = in1 + ilm2e;
							cgh = ghit(0, 0, nbl, l1, m1, l2, m2, c1, c1ao, c4, c6);
							cgk = ghit(0, 1, nbl, l1, m1, l2, m2, c1, c1ao, c4, c6);
							am[i1 - 1][i2 - 1] = cgh;
							am[i1 - 1][i2e - 1] = cgk;
							am[i1e - 1][i2 - 1] = cgk;
							am[i1e - 1][i2e - 1] = cgh;
							am[j1 - 1][j2 - 1] = cgh * (1.0 * ish);
							am[j1 - 1][j2e - 1] = cgk * (1.0 * isk);
							am[j1e - 1][j2 - 1] = cgk * (1.0 * isk);
							am[j1e - 1][j2e - 1] = cgh * (1.0 * ish);
						}
					}
				} // im1 loop
			} // l1 loop
		} // n2 loop
	} // n1 loop
	for (int n1 = 1; n1 <= c4->nsph; n1++) { // GPU portable?
		int in1 = (n1 - 1) * c4->nlim;
		for (int l1 = 1; l1 <= c4->li; l1++) {
			dm = c1->rmi[l1 - 1][n1 - 1];
			de = c1->rei[l1 - 1][n1 - 1];
			int l1po = l1 + 1;
			int il1 = l1po * l1;
			int l1tpo = l1po + l1;
			for (int im1 = 1; im1 <= l1tpo; im1++) {
				int m1 = im1 - l1po;
				int ilm1 = il1 + m1;
				int i1 = in1 + ilm1;
				int i1e = i1 + ndi;
				for (int ilm2 = 1; ilm2 <= c4->nlim; ilm2++) {
					int i2 = in1 + ilm2;
					int i2e = i2 + ndi;
					am[i1 - 1][i2 - 1] = cc0;
					am[i1 - 1][i2e - 1] = cc0;
					am[i1e - 1][i2 - 1] = cc0;
					am[i1e - 1][i2e - 1] = cc0;
				}
				am[i1 - 1][i1 - 1] = dm;
				am[i1e - 1][i1e - 1] = de;
			} // im1 loop
		} // l1 loop
	} // n1 loop
}

/*! \brief C++ porting of HJV
 *
 * \param exri: `double`
 * \param vk: `double`
 * \param jer: `int &`
 * \param lcalc: `int &`
 * \param arg: `complex\<double\> &`
 * \param c1: `C1 *`
 * \param c1ao: `C1_AddOns *`
 * \param c4: `C4 *`
 */
void hjv(
		double exri, double vk, int &jer, int &lcalc, std::complex<double> &arg,
		C1 *c1, C1_AddOns *c1ao, C4 *c4
) {
	int nsphmo = c4->nsph - 1;
	int lit = c4->li + c4->li;
	int lmt = c4->li + c4->le;
	const int rfj_size = (lit > lmt) ? lit : lmt;
	const int rfn_size = c4->litpo;
	double *rfj, *rfn;
	rfj = new double[rfj_size];
	rfn = new double[rfn_size];
	jer = 0;
	int ivhb = 0;
	for (int nf40 = 1; nf40 <= nsphmo; nf40++) { // GPU portable?
		int nfpo = nf40 + 1;
		for (int ns40 = nfpo; ns40 <= c4->nsph; ns40++) {
			double rx = c1->rxx[nf40 - 1] - c1->rxx[ns40 - 1];
			double ry = c1->ryy[nf40 - 1] - c1->ryy[ns40 - 1];
			double rz = c1->rzz[nf40 - 1] - c1->rzz[ns40 - 1];
			double rr = sqrt(rx * rx + ry * ry + rz * rz);
			double rarg = rr * vk * exri;
			arg = std::complex<double>(rarg, 0.0);
			rbf(lit, rarg, lcalc, rfj);
			if (lcalc < lit) {
				jer = 1;
				delete[] rfj;
				delete[] rfn;
				return;
			}
			rnf(lit, rarg, lcalc, rfn);
			if (lcalc < lit) {
				jer = 2;
				delete[] rfj;
				delete[] rfn;
				return;
			}
			for (int lpo38 = 1; lpo38 <= c4->litpo; lpo38++) {
				double rpart = rfj[lpo38 - 1];
				double ipart = rfn[lpo38 - 1];
				c1ao->vh[lpo38 + ivhb - 1] = std::complex<double>(rpart, ipart);
			}
			ivhb += c4->litpo;
		} // ns40 loop
	} // nf40 loop
	ivhb = 0;
	for (int nf50 = 1; nf50 <= c4->nsph; nf50++) {
		double rx = c1->rxx[nf50 - 1];
		double ry = c1->ryy[nf50 - 1];
		double rz = c1->rzz[nf50 - 1];
		if (!(rx == 0.0 && ry == 0.0 && rz == 0.0)) {
			double rr = sqrt(rx * rx + ry * ry + rz * rz);
			double rarg = rr * vk * exri;
			rbf(lmt, rarg, lcalc, rfj);
			if (lcalc < lmt) {
				jer = 3;
				delete[] rfj;
				delete[] rfn;
				return;
			}
			for (int lpo47 = 1; lpo47 <= c4->lmtpo; lpo47++) {
				c1ao->vj0[lpo47 + ivhb - 1] = rfj[lpo47 - 1];
			}
		}
		ivhb += c4->lmtpo;
	} // nf50 loop
	delete[] rfj;
	delete[] rfn;
}

/*! \brief C++ porting of LUCIN
 *
 * \param am: Matrix of complex.
 * \param nddmst: `const int`
 * \param n: `int`
 * \param ier: `int &`
 */
void lucin(std::complex<double> **am, const int nddmst, int n, int &ier) {
	/* NDDMST  FIRST DIMENSION OF AM AS DECLARED IN DIMENSION
	 *         STATEMENT.
	 * N       NUMBER OF ROWS IN AM.
	 * IER     IS REPLACED BY 1 FOR SINGULARITY.
	 */
	double *v = new double[nddmst];
	std::complex<double> ctemp, cfun;
	std::complex<double> cc0 = std::complex<double>(0.0, 0.0);
	ier = 0;
	int nminus = n - 1;
	for (int i = 1; i <= n; i++) {
		double sum = 0.0;
		for (int j = 1; j <= n; j++) {
			sum += (
					am[i - 1][j - 1].real() * am[i - 1][j - 1].real()
					+ am[i - 1][j - 1].imag() * am[i - 1][j - 1].imag()
			);
		} // j1319 loop
		v[i - 1] = 1.0 / sum;
	} // i1309 loop
	// 2. REPLACE AM BY TRIANGULAR MATRICES (L,U) WHERE AM=L*U.
	//    REPLACE L(I,I) BY 1/L(I,I), READY FOR SECTION 4.
	//    (ROW INTERCHANGES TAKE PLACE, AND THE INDICES OF THE PIVOTAL ROWS
	//    ARE PLACED IN V.)
	/* >>> THERE APPEARS TO BE A BUG IN THE FOLLOWING LOOP <<< */
	for (int k = 1; k <= n; k++) {
		int kplus = k + 1;
		int kminus = k - 1;
		int l = k;
		double psqmax = 0.0;
		for (int i = k; i <= n; i++) {
			cfun = cdtp(-am[i - 1][k - 1], am, i, 1, k, kminus);
			ctemp = -cfun;
			am[i - 1][k - 1] = ctemp;
			double psq = v[i - 1] * (ctemp.real() * ctemp.real() + ctemp.imag() * ctemp.imag());
			if (psq > psqmax) {
				psqmax = psq;
				l = i;
			}
		} // i2029 loop
		if (l != k) {
			for (int j = 1; j <= n; j++) {
				ctemp = am[k - 1][j - 1];
				am[k - 1][j - 1] = am[l - 1][j - 1];
				am[l - 1][j - 1] = ctemp;
			} // j2049 loop
			v[l - 1] = v[k - 1];
		}
		// label 2011
		v[k - 1] = 1.0 * l;
		if (psqmax == 0.0) {
			ier = 1;
			delete[] v;
			return;
		}
		ctemp = 1.0 / am[k - 1][k - 1];
		am[k - 1][k - 1] = ctemp;
		if (kplus <= n) {
			for (int j = kplus; j <= n; j++) {
				cfun = cdtp(-am[k - 1][j - 1], am, k, 1, j, kminus);
				am[k - 1][j - 1] = -ctemp * cfun;
			} // j2059 loop
		}
	} // k2019 loop
	// 4.  REPLACE AM BY ITS INVERSE AMINV.
	// 4.1 REPLACE L AND U BY THEIR INVERSE LINV AND UINV.
	for (int k = 1; k <= nminus; k++) {
		int kplus = k + 1;
		for (int i = kplus; i <= n; i++) {
			cfun = cdtp(cc0, am, i, k, k, i - k);
			am[i - 1][k - 1] = -am[i - 1][i - 1] * cfun;
			cfun = cdtp(am[k - 1][i - 1], am, k, kplus, i, i - k - 1);
			am[k - 1][i - 1] = -cfun;
		} // i4119 loop
	} // k4109 loop
	// 4.2 FORM AMINV=UINV*LINV.
	for (int k = 1; k <= n; k++) {
		for (int i = 1; i <= n; i++) {
			if (i < k) {
				cfun = cdtp(cc0, am, i, k, k, n - k + 1);
				am[i - 1][k -1] = cfun;
			}
			else {
				cfun = cdtp(am[i - 1][k - 1], am, i, i + 1, k, n - i);
				am[i - 1][k - 1] = cfun;
			}
		} // i4119 loop
	} // k4209 loop
	// 4.3 INTERCHANGE COLUMNS OF AMINV AS SPECIFIED BY V, BUT IN REVERSE
	//     ORDER.
	for (int l = 1; l <= n; l++) {
		int k = n - l + 1;
		int kcol = (int)(v[k - 1]);
		if (kcol != k) {
			for (int i = 1; i <= n; i++) {
				ctemp = am[i - 1][k - 1];
				am[i - 1][k - 1] = am[i - 1][kcol - 1];
				am[i - 1][kcol - 1] = ctemp;
			} // i4319 loop
		}
	} // l4309 loop
	delete[] v;
}

/*! \brief C++ porting of POLAR
 *
 * \param x: `double`
 * \param y: `double`
 * \param z: `double`
 * \param r: `double &`
 * \param cth: `double &`
 * \param sth: `double &`
 * \param cph: `double &`
 * \param sph: `double &`
 */
void polar(
		double x, double y, double z, double &r, double &cth, double &sth,
		double &cph, double &sph
) {
	bool onx = (y == 0.0);
	bool ony = (x == 0.0);
	bool onz = (onx && ony);
	double rho = 0.0;
	if (!onz) {
		if (!onx) {
			if (!ony) {
				rho = sqrt(x * x + y * y);
				cph = x / rho;
				sph = y / rho;
				// goes to 25
			} else { // label 20
				rho = (y > 0.0) ? y : -y;
				cph = 0.0;
				sph = (y > 0.0) ? 1.0 : -1.0;
				// goes to 25
			}
		} else { // label 15
			rho = (x > 0.0) ? x : -x;
			cph = (x > 0.0) ? 1.0 : -1.0;
			sph = 0.0;
			// goes to 25
		}
	} else { // label 10
		cph = 1.0;
		sph = 0.0;
		// goes to 25
	}
	// label 25
	if (z == 0.0) {
		if (!onz) {
			r = rho;
			cth = 0.0;
			sth = 1.0;
			// returns
		} else { // label 30
			r = 0.0;
			cth = 1.0;
			sth = 0.0;
			// returns
		}
	} else { // label 35
		if (!onz) {
			r = sqrt(rho * rho + z * z);
			cth = z / r;
			sth = rho / r;
			// returns
		} else { // label 40
			r = (z > 0.0) ? z : -z;
			cth = (z > 0.0) ? 1.0 : -1.0;
			sth = 0.0;
			// returns
		}
	}
}

/*! \brief C++ porting of R3J000
 *
 * \param j2: `int`
 * \param j3: `int`
 * \param c6: `C6 *` Pointer to a C6 instance.
 */
void r3j000(int j2, int j3, C6 *c6) {
	int jmx = j3 + j2;
	if (jmx <= 0) {
		c6->rac3j[0] = 1.0;
		return;
	}
	int jmn = j3 - j2;
	if (jmn < 0) jmn *= -1;
	int njmo = (jmx - jmn) / 2;
	int jf = jmx + jmx + 1;
	int isn = 1;
	if (jmn % 2 != 0) isn = -1;
	if (njmo <= 0) {
		double sj = 1.0 * jf;
		double cnr = (1 / sqrt(sj)) * isn;
		c6->rac3j[0] = cnr;
		return;
	}
	double sjr = 1.0 * jf;
	int jmxpos = (jmx + 1) * (jmx + 1);
	int jmns = jmn * jmn;
	int j1mo = jmx - 1;
	int j1s = (j1mo + 1) * (j1mo + 1);
	double cj = sqrt(1.0 * (jmxpos - j1s) * (j1s - jmns));
	int j1mos = j1mo * j1mo;
	double cjmo = sqrt(1.0 * (jmxpos - j1mos) * (j1mos - jmns));
	if (njmo <= 1) {
		c6->rac3j[0] = -cj / cjmo;
		double sj = sjr + (c6->rac3j[0] * c6->rac3j[0]) * (jf - 4);
		double cnr = (1.0 / sqrt(sj)) * isn;
		c6->rac3j[1] = cnr;
		c6->rac3j[0] *= cnr;
		return;
	}
	int nj = njmo + 1;
	int nmat = (nj + 1) / 2;
	c6->rac3j[nj - 1] = 1.0;
	c6->rac3j[njmo - 1] = -cj / cjmo;
	if (nmat != njmo) {
		int nbr = njmo - nmat;
		for (int ibr45 = 1; ibr45 <= nbr; ibr45++) {
			int irr = nj - ibr45;
			jf -= 4;
			j1mo -= 2;
			j1s = (j1mo + 1) * (j1mo + 1);
			cj = sqrt(1.0 * (jmxpos - j1s) * (j1s - jmns));
			j1mos = j1mo * j1mo;
			cjmo = sqrt(1.0 * (jmxpos - j1mos) * (j1mos - jmns));
			c6->rac3j[irr - 2] = c6->rac3j[irr - 1] * (-cj / cjmo);
			sjr = sjr + (c6->rac3j[irr - 1] * c6->rac3j[irr - 1]) * jf;
		}
	}
	// label 50
	double racmat = c6->rac3j[nmat - 1];
	sjr = sjr + (racmat * racmat) * (jf - 4);
	c6->rac3j[0] = 1.0;
	jf = jmn + jmn + 1;
	double sjl = 1.0 * jf;
	int j1pt = jmn + 2;
	int j1pos = (j1pt - 1) * (j1pt - 1);
	double cjpo = sqrt(1.0 * (jmxpos - j1pos) * (j1pos - jmns));
	int j1pts = j1pt * j1pt;
	double cjpt = sqrt(1.0 * (jmxpos - j1pts) * (j1pts - jmns));
	c6->rac3j[1] = -cjpo / cjpt;
	int nmatmo = nmat - 1;
	if (nmatmo >= 2) {
		for (int irl70 = 2; irl70 <= nmatmo; irl70++) {
			jf += 4;
			j1pt += 2;
			j1pos = (j1pt - 1) * (j1pt - 1);
			cjpo = sqrt(1.0 * (jmxpos - j1pos) * (j1pos - jmns));
			j1pts = j1pt * j1pt;
			cjpt = sqrt(1.0 * (jmxpos - j1pts) * (j1pts - jmns));
			c6->rac3j[irl70] = c6->rac3j[irl70 - 1] * (-cjpo / cjpt);
			sjl = sjl + (c6->rac3j[irl70 - 1] * c6->rac3j[irl70 - 1]) * jf;
		}
	}
	// label 75
	double ratrac = racmat / c6->rac3j[nmat - 1];
	double rats = ratrac * ratrac;
	double sj = sjr + sjl * rats;
	c6->rac3j[nmat - 1] = racmat;
	double cnr = (1.0 / sqrt(sj)) * isn;
	for (int irr80 = nmat; irr80 <= nj; irr80++) {
		c6->rac3j[irr80 - 1] *= cnr;
	}
	double cnl = cnr * ratrac;
	for (int irl85 = 1; irl85 <= nmatmo; irl85++) {
		c6->rac3j[irl85 - 1] *= cnl;
	}
}

/*! \brief C++ porting of SCR0
 *
 * \param vk: `double` QUESTION: definition?
 * \param exri: `double` External medium refractive index. QUESTION: correct?
 * \param c1: `C1 *` Pointer to a C1 instance.
 * \param c1ao: `C1_AddOns *` Pointer to C1_AddOns instance.
 * \param c3: `C3 *` Pointer to a C3 instance.
 * \param c4: `C4 *` Pointer to a C4 structure.
 */
void scr0(double vk, double exri, C1 *c1, C1_AddOns *c1ao, C3 *c3, C4 * c4) {
	const std::complex<double> cc0(0.0, 0.0);
	double exdc = exri * exri;
	double ccs = 4.0 * acos(0.0) / (vk * vk);
	double cccs = ccs / exdc;
	std::complex<double> sum21, rm, re, csam;
	csam = -(ccs / (exri * vk)) * std::complex<double>(0.0, 0.5);
	double scs = 0.0, ecs = 0.0, acs = 0.0;
	c3->tfsas = cc0;
	for (int i14 = 1; i14 <= c4->nsph; i14++) {
		int iogi = c1->iog[i14 - 1];
		if (iogi >= i14) {
			double sums = 0.0;
			sum21 = cc0;
			for (int l10 = 1; l10 <= c4->li; l10++) {
				double fl = 1.0 * (l10 + l10 + 1);
				rm = 1.0 / c1->rmi[l10 - 1][i14 - 1];
				re = 1.0 / c1->rei[l10 - 1][i14 - 1];
				double rvalue = (dconjg(rm) * rm + dconjg(re) * re).real() * fl;
				sums += rvalue;
				sum21 += ((rm + re) * fl);
			} // l10 loop
			sum21 *= -1.0;
			double scasec = cccs * sums;
			double extsec = -cccs * sum21.real();
			double abssec = extsec - scasec;
			c1->sscs[i14 - 1] = scasec;
			c1->sexs[i14 - 1] = extsec;
			c1->sabs[i14 - 1] = abssec;
			double gcss = c1->gcsv[i14 - 1];
			c1->sqscs[i14 - 1] = scasec / gcss;
			c1->sqexs[i14 - 1] = extsec / gcss;
			c1->sqabs[i14 - 1] = abssec / gcss;
			c1->fsas[i14 - 1] = sum21 * csam;
		}
		// label 12
		scs += c1->sscs[iogi - 1];
		ecs += c1->sexs[iogi - 1];
		acs += c1->sabs[iogi - 1];
		c3->tfsas += c1->fsas[iogi - 1];
	} // i14 loop
}

/*! \brief C++ porting of STR
 *
 * \param rcf: `double **` Matrix of double.
 * \param c1: `C1 *` Pointer to a C1 instance.
 * \param c1ao: `C1_AddOns *` Pointer to C1_AddOns instance.
 * \param c3: `C3 *` Pointer to a C3 instance.
 * \param c4: `C4 *` Pointer to a C4 structure.
 * \param c6: `C6 *` Pointer to a C6 instance.
 */
void str(double **rcf, C1 *c1, C1_AddOns *c1ao, C3 *c3, C4 *c4, C6 *c6) {
	std::complex<double> *ylm;
	const double pi = acos(-1.0);
	c3->gcs = 0.0;
	for (int i18 = 1; i18 <= c4->nsph; i18++) {
		int iogi = c1->iog[i18 - 1];
		if (iogi >= i18) {
			double gcss = pi * c1->ros[i18 - 1] * c1->ros[i18 - 1];
			c1->gcsv[i18 - 1] = gcss;
			int nsh = c1->nshl[i18 - 1];
			for (int j16 = 1; j16 <= nsh; j16++) {
				c1->rc[i18 - 1][j16 - 1] = rcf[i18 - 1][j16 - 1] * c1->ros[i18 - 1];
			} // j16 loop
			c3->gcs += gcss;
		}
	} // i18 loop
	int ylm_size = (c4->litpos > c4->lmtpos) ? c4->litpos : c4->lmtpos;
	ylm = new std::complex<double>[ylm_size]();
	int i = 0;
	for (int l1po28 = 1; l1po28 <= c4->lmpo; l1po28++) {
		int l1 = l1po28 - 1;
		for (int l2 = 1; l2 <= c4->lm; l2++) {
			r3j000(l1, l2, c6);
			c1ao->ind3j[l1po28 - 1][l2 - 1] = i;
			int lmnpo = (l2 > l1) ? l2 - l1 + 1 : l1 - l2 + 1;
			int lmxpo = l2 + l1 + 1;
			int lpo28 = lmnpo;
			int il = 0;
			while (lpo28 <= lmxpo) {
				i++;
				il++;
				c1ao->v3j0[i - 1] = c6->rac3j[il - 1];
				lpo28 += 2;
			}
		} // l2 loop
	} // l1po28 loop
	int nsphmo = c4->nsph - 1;
	int lit = c4->li + c4->li;
	int ivy = 0;
	for (int nf40 = 1; nf40 <= nsphmo; nf40++) { // GPU portable?
		int nfpo = nf40 + 1;
		for (int ns40 = nfpo; ns40 <= c4->nsph; ns40++) {
			double rx = c1->rxx[nf40 - 1] - c1->rxx[ns40 - 1];
			double ry = c1->ryy[nf40 - 1] - c1->ryy[ns40 - 1];
			double rz = c1->rzz[nf40 - 1] - c1->rzz[ns40 - 1];
			double rr = 0.0;
			double crth = 0.0, srth = 0.0, crph = 0.0, srph = 0.0;
			polar(rx, ry, rz, rr, crth, srth, crph, srph);
			sphar(crth, srth, crph, srph, lit, ylm);
			for (int iv38 = 1; iv38 <= c4->litpos; iv38++) {
				c1ao->vyhj[iv38 + ivy - 1] = dconjg(ylm[iv38 - 1]);
			} // iv38 loop
			ivy += c4->litpos;
		} // ns40 loop
	} // nf40 loop
	int lmt = c4->li + c4->le;
	ivy = 0;
	for (int nf50 = 1; nf50 <= c4->nsph; nf50++) {
		double rx = c1->rxx[nf50 - 1];
		double ry = c1->ryy[nf50 - 1];
		double rz = c1->rzz[nf50 - 1];
		if (rx != 0.0 || ry != 0.0 || rz != 0.0) {
			double rr = 0.0;
			double crth = 0.0, srth = 0.0, crph = 0.0, srph = 0.0;
			polar(rx, ry, rz, rr, crth, srth, crph, srph);
			sphar(crth, srth, crph, srph, lmt, ylm);
			for (int iv48 = 1; iv48 <= c4->lmtpos; iv48++) {
				c1ao->vyj0[iv48 + ivy - 1] = dconjg(ylm[iv48 - 1]);
			} // iv48 loop
		}
		ivy += c4->lmtpos;
	} // nf50 loop
	delete[] ylm;
}

/*! \brief C++ porting of ZTM
 *
 * \param am: Matrix of complex.
 * \param c1: `C1 *` Pointer to a C1 instance.
 * \param c1ao: `C1_AddOns *` Pointer to C1_AddOns instance.
 * \param c4: `C4 *` Pointer to a C4 structure.
 * \param c6: `C6 *` Pointer to a C6 instance.
 * \param c9: `C9 *` Pointer to a C9 instance.
 */
void ztm(std::complex<double> **am, C1 *c1, C1_AddOns *c1ao, C4 *c4, C6 *c6, C9 * c9) {
	std::complex<double> gie, gle, a1, a2, a3, a4, sum1, sum2, sum3, sum4;
	const std::complex<double> cc0(0.0, 0.0);
	int ndi = c4->nsph * c4->nlim;
	int i2 = 0;
	for (int n2 = 1; n2 <= c4->nsph; n2++) { // GPU portable?
		for (int l2 = 1; l2 <= c4->li; l2++) {
			int l2tpo = l2 + l2 + 1;
			int m2 = -l2 - 1;
			for (int im2 = 1; im2 <= l2tpo; im2++) {
				m2++;
				i2++;
				int i3 = 0;
				for (int l3 = 1; l3 <= c4->le; l3++) {
					int l3tpo = l3 + l3 + 1;
					int m3 = -l3 - 1;
					for (int im3 = 1; im3 <= l3tpo; im3++) {
						m3++;
						i3++;
						c9->gis[i2 - 1][i3 - 1] = ghit(2, 0, n2, l2, m2, l3, m3, c1, c1ao, c4, c6);
						c9->gls[i2 - 1][i3 - 1] = ghit(2, 1, n2, l2, m2, l3, m3, c1, c1ao, c4, c6);
					} // im3 loop
				} // l3 loop
			} // im2 loop
		} // l2 loop
	} // n2 loop
	/* // DEBUG CODE
	std::complex<double> dbgtst(0.0, 0.0);
	for (int di = 0; di < ndi; di++) {
		for (int dj = 0; dj < c4->nlem; dj++) dbgtst += c9->gis[di][dj];
	}
	printf("DEBUG: in ZTM init sgis = (%lE, %lE)\n", dbgtst.real(), dbgtst.imag());
	dbgtst = std::complex<double>(0.0, 0.0);
	for (int di = 0; di < ndi; di++) {
		for (int dj = 0; dj < c4->nlem; dj++) dbgtst += c9->gls[di][dj];
	}
	printf("DEBUG: in ZTM init sgls = (%lE, %lE)\n", dbgtst.real(), dbgtst.imag());
	// END DEBUG */
	for (int i1 = 1; i1 <= ndi; i1++) { // GPU portable?
		int i1e = i1 + ndi;
		for (int i3 = 1; i3 <= c4->nlem; i3++) {
			int i3e = i3 + c4->nlem;
			sum1 = cc0;
			sum2 = cc0;
			sum3 = cc0;
			sum4 = cc0;
			for (int i2 = 1; i2 <= ndi; i2++) {
				int i2e = i2 + ndi;
				gie = c9->gis[i2 - 1][i3 - 1];
				gle = c9->gls[i2 - 1][i3 - 1];
				a1 = am[i1 - 1][i2 - 1];
				a2 = am[i1 - 1][i2e - 1];
				a3 = am[i1e - 1][i2 - 1];
				a4 = am[i1e - 1][i2e - 1];
				sum1 += (a1 * gie + a2 * gle);
				sum2 += (a1 * gle + a2 * gie);
				sum3 += (a3 * gie + a4 * gle);
				sum4 += (a3 * gle + a4 * gie);
			} // i2 loop
			c9->sam[i1 - 1][i3 - 1] = sum1;
			c9->sam[i1 - 1][i3e - 1] = sum2;
			c9->sam[i1e - 1][i3 - 1] = sum3;
			c9->sam[i1e - 1][i3e - 1] = sum4;
		} // i3 loop
	} // i1 loop
	for (int i1 = 1; i1 <= ndi; i1++) {
		for (int i0 = 1; i0 <= c4->nlem; i0++) {
			c9->gis[i1 - 1][i0 - 1] = dconjg(c9->gis[i1 - 1][i0 - 1]);
			c9->gls[i1 - 1][i0 - 1] = dconjg(c9->gls[i1 - 1][i0 - 1]);
		} // i0 loop
	} // i1 loop
	int nlemt = c4->nlem + c4->nlem;
	for (int i0 = 1; i0 <= c4->nlem; i0++) {
		int i0e = i0 + c4->nlem;
		for (int i3 = 1; i3 <= nlemt; i3++) {
			sum1 = cc0;
			sum2 = cc0;
			for (int i1 = 1; i1 <= ndi; i1 ++) {
				int i1e = i1 + ndi;
				a1 = c9->sam[i1 - 1][i3 - 1];
				a2 = c9->sam[i1e - 1][i3 - 1];
				gie = c9->gis[i1 - 1][i0 - 1];
				gle = c9->gls[i1 - 1][i0 - 1];
				sum1 += (a1 * gie + a2 * gle);
				sum2 += (a1 * gle + a2 * gie);
			} // i1 loop
			c1ao->am0m[i0 - 1][i3 - 1] = -sum1;
			c1ao->am0m[i0e - 1][i3 - 1] = -sum2;
		} // i3 loop
	} // i0 loop
}

/*! \brief Write a matrix to a log file (debug function).
 *
 *  \param mat: Matrix of complex.
 *  \param rows: `int`
 *  \param cols: `int`
 */
void logmat(std::complex<double> **mat, int rows, int cols) {
	FILE* logfile = fopen("c_matrix.log", "w");
	for (int i = 0; i < rows; i++) {
		for (int j = 0; j < cols; j++) {
			fprintf(logfile, "R:%3d,%3d,%15.7lE\n", i + 1, j + 1, mat[i][j].real());
			fprintf(logfile, "I:%3d,%3d,%15.7lE\n", i + 1, j + 1, mat[i][j].imag());
		}
	}
	fclose(logfile);
}

/*! \brief Sum all the elements of a matrix (debug function).
 *
 *  \param mat: Matrix of complex.
 *  \param rows: `int`
 *  \param cols: `int`
 */
std::complex<double> summat(std::complex<double> **mat, int rows, int cols) {
	std::complex<double> result(0.0, 0.0);
	for (int i = 0; i < rows; i++) {
		for (int j = 0; j < cols; j++) result += mat[i][j];
	}
	return result;
}

#endif
